import logging
import os
import re
import secrets
import shutil
import socket
import subprocess
import tempfile
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import psycopg2  # type: ignore
import yaml

import balsam.server

SERVER_INFO_FILENAME = "server-info.yml"
MIN_VERSION = (10, 0, 0)
POSTGRES_MSG = f"""You need PostgreSQL {'.'.join(map(str, MIN_VERSION))} or newer.
Ensure pg_ctl is in the search PATH, and double-check version with pg_ctl --version.
You can easily grab the Postgres binaries at:
https://www.enterprisedb.com/download-postgresql-binaries
"""

logger = logging.getLogger(__name__)
OperationalError = psycopg2.OperationalError


def make_token() -> str:
    return secrets.token_urlsafe()


def make_token_file() -> Tuple[str, str]:
    with tempfile.NamedTemporaryFile(delete=False, mode="w", encoding="utf-8") as fp:
        token = make_token()
        fp.write(token)
        tmp_pwfile = fp.name
    return token, tmp_pwfile


def identify_free_port() -> int:
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("", 0))
    port = int(sock.getsockname()[1])
    sock.close()
    return port


def identify_hostport() -> Tuple[str, int]:
    host = socket.gethostname()
    port = identify_free_port()
    return (host, port)


def write_pwfile(
    db_path: Union[Path, str],
    username: str,
    password: str,
    host: str,
    port: int,
    database: str,
    filename: str = SERVER_INFO_FILENAME,
) -> Dict[str, Any]:
    pw_dict = dict(username=username, password=password, host=host, port=port, database=database)
    pw_dict["scheme"] = "postgres"
    pw_dict["client_class"] = "balsam.client.DirectAPIClient"
    with open(Path(db_path).joinpath(filename), "w") as fp:
        yaml.dump(pw_dict, fp, sort_keys=False, indent=4)
    return pw_dict


def load_pwfile(db_path: Union[Path, str], filename: str = SERVER_INFO_FILENAME) -> Dict[str, Any]:
    with open(Path(db_path).joinpath(filename)) as fp:
        pw_dict = yaml.safe_load(fp)
    assert isinstance(pw_dict, dict)
    return pw_dict


def load_dsn(db_path: Union[Path, str], filename: str = SERVER_INFO_FILENAME) -> str:
    pw_dict = load_pwfile(db_path, filename)
    user = pw_dict["username"]
    pwd = pw_dict["password"]
    host = pw_dict["host"]
    port = pw_dict["port"]
    db = pw_dict["database"]
    return f"postgresql://{user}:{pwd}@{host}:{port}/{db}"


def create_new_db(db_path: Union[Path, str] = "balsamdb", database: str = "balsam") -> Dict[str, Any]:
    """
    Create & start a new PostgresDB cluster inside `db_path`
    A DB named `database` is created. If `pwfile` given, write DB credentials to
    `db_path/pwfile`.  Returns a dict containing credentials/connection
    info.
    """
    version_check()

    superuser, password = init_db_cluster(db_path)
    host, port = identify_hostport()
    mutate_conf_port(str(db_path), port)

    pw_dict = write_pwfile(
        db_path=db_path,
        username=superuser,
        password=password,
        host=host,
        port=port,
        database=database,
    )

    start_db(str(db_path))
    create_database(new_dbname=database, **pw_dict)
    return pw_dict


def configure_balsam_server(username: str, password: str, host: str, port: int, database: str, **kwargs: Any) -> str:
    dsn = f"postgresql://{username}:{password}@{host}:{port}/{database}"
    os.environ["balsam_database_url"] = dsn
    balsam.server.settings.database_url = dsn
    return dsn


def configure_balsam_server_from_dsn(dsn: str) -> None:
    os.environ["balsam_database_url"] = dsn
    os.environ["BALSAM_DATABASE_URL"] = dsn
    balsam.server.settings.database_url = dsn


def run_alembic_migrations(dsn: str, downgrade: Any = None) -> None:
    from alembic import command  # type: ignore
    from alembic.config import Config  # type: ignore

    import balsam.server.models.alembic as alembic

    migrations_path = str(Path(alembic.__file__).parent)

    logger.info(f"Running DB migrations in {migrations_path}")
    alembic_cfg = Config()
    alembic_cfg.set_main_option("script_location", str(migrations_path))
    # alembic_cfg.set_main_option("sqlalchemy.url", dsn)
    if downgrade is None:
        command.upgrade(alembic_cfg, "head")
    else:
        command.downgrade(alembic_cfg, downgrade)


# *******************************
# Functions that use a subprocess
# *******************************
def version_check() -> bool:
    p = subprocess.run(
        "pg_ctl --version",
        shell=True,
        encoding="utf-8",
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        check=True,
    )
    stdout = p.stdout.strip()
    pattern = re.compile(r"(\d+\.)?(\d+\.)?(\*|\d+)")
    search_result = pattern.search(stdout)
    if not search_result:
        raise RuntimeError(f"Could not parse Postgres version from `pg_ctl --version`: {stdout}")
    match = search_result.group()
    major, minor, *rest = match.split(".")
    version_info: Tuple[int, ...]
    if rest and int(rest[0]) > 0:
        version_info = (int(major), int(minor), int(rest[0]))
    else:
        version_info = (int(major), int(minor))
    if version_info >= MIN_VERSION:
        return True
    raise RuntimeError(f"PostgreSQL {version_info} too old. {POSTGRES_MSG}")


def mutate_conf_port(db_path: Union[str, Path], port: int) -> None:
    settings_path = Path(db_path) / Path("postgresql.conf")
    if not settings_path.exists():
        raise ValueError(f"nonexistant {settings_path}")

    new_port = f"port={port}"
    cmd = f"sed -i -E 's/^[[:space:]]*port.*/{new_port}/g' {settings_path}"
    subprocess.run(cmd, shell=True, check=True)


def init_db_cluster(db_path: Union[Path, str], superuser: str = "postgres") -> Tuple[str, str]:
    db_path = Path(db_path)
    if db_path.exists():
        raise ValueError(f"{db_path} already exists")

    token, tmp_pwfile = make_token_file()

    subprocess.run(
        f"initdb -D {db_path} --auth scram-sha-256 -U {superuser} --pwfile {tmp_pwfile}",
        shell=True,
        check=True,
    )
    shutil.move(tmp_pwfile, db_path.joinpath("pwfile").as_posix())

    with open(db_path / "postgresql.conf", "a") as fp:
        fp.write("listen_addresses = '*' # appended from balsam init\n")
        fp.write("port=0 # appended from balsam init\n")
        fp.write("password_encryption = 'scram-sha-256'\n")
        fp.write("max_connections=210 # appended from balsam init\n")
        fp.write("shared_buffers=2GB # appended from balsam init\n")
        fp.write("synchronous_commit=off # appended from balsam init\n")
        fp.write("wal_writer_delay=400ms # appended from balsam init\n")
        fp.write("client_encoding = 'UTF8'\n")
        fp.write("timezone = 'UTC'\n")

        # logging
        fp.write("logging_collector=on # appended from balsam init\n")
        fp.write("log_min_duration_statement=0 # appended from balsam init\n")
        fp.write("log_connections=on # appended from balsam init\n")
        fp.write("log_duration=on # appended from balsam init\n")

    with open(db_path / "pg_hba.conf", "a") as fp:
        fp.write("host all all ::1/128 scram-sha-256\n")
        fp.write("host all all 0.0.0.0/0 scram-sha-256\n")
    return superuser, token


def start_db(db_path: Union[Path, str]) -> None:
    log_path = Path(db_path) / "postgres.log"
    start_cmd = f"pg_ctl -w start -D {db_path} -l {log_path} --mode=smart"
    logger.info(start_cmd)
    subprocess.run(start_cmd, shell=True, check=True)


def stop_db(db_path: Union[Path, str]) -> None:
    stop_cmd = f"pg_ctl -w stop -D {db_path} --mode=smart"
    res = subprocess.run(stop_cmd, shell=True)
    if res.returncode != 0:
        raise RuntimeError("Could not send stop signal; is the DB running on a different host?")


# ***********************************
# Functions that use a psycopg cursor
# ***********************************
class PgCursor:
    def __init__(
        self, host: str, port: int, username: str, password: str, database: str = "balsam", autocommit: bool = False
    ) -> None:
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.database = database
        self.autocommit = autocommit

    def __enter__(self) -> Any:
        self.conn = psycopg2.connect(
            dbname=self.database,
            user=self.username,
            password=self.password,
            host=self.host,
            port=self.port,
        )
        self.conn.autocommit = self.autocommit
        self.cursor = self.conn.cursor()
        return self.cursor

    def __exit__(self, exc_type: Any, value: Any, traceback: Any) -> None:
        if not self.autocommit:
            if exc_type is None:
                self.conn.commit()
            else:
                self.conn.rollback()
        self.cursor.close()
        self.conn.close()


def create_database(new_dbname: str, host: str, port: int, username: str, password: str, **kwargs: Any) -> None:
    with PgCursor(host, port, username, password, database="", autocommit=True) as cur:
        cur.execute(
            f"""
        CREATE DATABASE {new_dbname};
        """
        )


def connections_list(
    host: str, port: int, username: str, password: str, database: str = "balsam", **kwargs: Any
) -> List[str]:
    with PgCursor(host, port, username, password) as cur:
        cur.execute(
            f"""
        SELECT pid,application_name,usename,state,substr(query, 1, 60) \
        FROM pg_stat_activity WHERE datname = '{database}';
        """
        )
        conns = cur.fetchall()
        return list(conns)


def create_user_and_pwfile(
    new_user: str,
    host: str,
    port: int,
    username: str,
    password: str,
    database: str = "balsam",
    pwfile: Optional[str] = None,
) -> None:
    """
    Create new Postgres user `new_user` and auto-generated
    token for the database at `host:port`.  A pwfile
    containing the DB credentials is written for transfer to the new user
    """
    token = make_token()
    pg_add_user(new_user, token, host, port, username, password, database)

    if pwfile is None:
        pwfile = f"{new_user}-server-info.yml"
    write_pwfile(
        db_path=".",
        username=new_user,
        password=token,
        host=host,
        port=port,
        database=database,
        filename=pwfile,
    )

    print(
        f"New Postgres user data in {pwfile}.\n"
        f"Protect this file in transit to {new_user}! "
        f"It contains the token necessary to reach the DB. "
        f"The new user can then reach the DB with "
        f"`balsam login --file {pwfile}`"
    )


def pg_add_user(
    new_user: str, new_password: str, host: str, port: int, username: str, password: str, database: str = "balsam"
) -> None:
    with PgCursor(host, port, username, password) as cur:
        cur.execute(
            f"""
        CREATE ROLE {new_user} WITH LOGIN PASSWORD '{new_password}'; \
        grant all privileges on database {database} to {new_user}; \
        GRANT ALL PRIVILEGES ON ALL TABLES IN SCHEMA public TO {new_user}; \
        GRANT ALL PRIVILEGES ON ALL SEQUENCES IN SCHEMA public TO {new_user};
        """
        )


def drop_user(
    deleting_user: str, host: str, port: int, username: str, password: str, database: str = "balsam", **kwargs: Any
) -> None:
    with PgCursor(host, port, username, password) as cur:
        cur.execute(
            f"""
        REVOKE ALL privileges on all tables in schema public FROM {deleting_user}; \
        REVOKE ALL privileges on all sequences in schema public FROM {deleting_user}; \
        REVOKE ALL privileges on database {database} FROM {deleting_user}; \
        DROP ROLE {deleting_user};
        """
        )


def list_users(host: str, port: int, username: str, password: str, **kwargs: Any) -> List[str]:
    with PgCursor(host, port, username, password) as cur:
        cur.execute("SELECT usename FROM pg_catalog.pg_user;")
        user_list = cur.fetchall()
        return list(user_list)



import base64
import logging
from functools import lru_cache
from typing import Any, Optional

import dill  # type: ignore
import dill.source  # type: ignore
from tblib import Traceback  # type: ignore

logger = logging.getLogger(__name__)


class EmptyPayload(ValueError):
    pass


class SerializeError(ValueError):
    pass


class DeserializeError(ValueError):
    pass


def get_source(obj: Any) -> str:
    source: str = dill.source.getsource(obj, lstrip=True)
    return source


def _serialize(obj: Any) -> str:
    dump: bytes = dill.dumps(obj, recurse=True)
    return base64.b64encode(dump).decode("utf-8")


_cached_serialize = lru_cache(256)(_serialize)


def serialize(obj: Any) -> str:
    try:
        try:
            return _cached_serialize(obj)
        except TypeError:
            return _serialize(obj)
    except Exception as exc:
        logger.exception(f"Failed to serialize {obj}")
        raise SerializeError(str(exc)) from exc


@lru_cache(256)
def deserialize(payload: str) -> Any:
    if not payload:
        raise EmptyPayload
    try:
        decoded = base64.b64decode(payload)
        return dill.loads(decoded)
    except Exception as exc:
        logger.exception("Failed to deserialize")
        raise DeserializeError(str(exc)) from exc


# See RemoteExceptionWrapper in parsl.apps.errors
# https://github.com/Parsl/parsl/blob/master/parsl/app/errors.py
def serialize_exception(exc: Exception) -> str:
    if exc.__traceback__ is not None:
        tb = Traceback(exc.__traceback__)
    else:
        tb = None
    return serialize((exc, tb))


def raise_from_serialized(payload: str) -> None:
    exc: Exception
    tb: Optional[Traceback]

    try:
        exc, tb = deserialize(payload)
    except DeserializeError as deser_exc:
        logger.error(f"An exception was transmitted, but it could not be unpacked here due to: {deser_exc}")
        logger.error("You may find the original error in the job.out file of the Job's working directory")
        raise

    if tb is not None:
        raise exc.with_traceback(tb.as_traceback())
    else:
        raise exc



from .base import Base, create_tables, get_engine, get_session
from .tables import App, BatchJob, Job, LogEvent, Session, Site, TransferItem, User, job_deps

__all__ = [
    "Base",
    "create_tables",
    "get_engine",
    "get_session",
    "App",
    "BatchJob",
    "Job",
    "LogEvent",
    "Session",
    "Site",
    "TransferItem",
    "User",
    "job_deps",
]



import logging
import secrets
import string
from datetime import datetime
from typing import Any, Dict
from uuid import UUID

from fastapi import APIRouter, Depends, Form, HTTPException, Request
from sqlalchemy.orm import Session, exc

from balsam.schemas import UserOut
from balsam.server import settings
from balsam.server.models import get_session
from balsam.server.models.crud import users

from .token import create_access_token

logger = logging.getLogger(__name__)

VERIFICATION_PATH = "/auth/oauth/ALCF/login/device"
GRANT_TYPE = "urn:ietf:params:oauth:grant-type:device_code"

router = APIRouter(prefix="/device")


def generate_device_code() -> str:
    """
    Returns a high-entropy device code (never typed or copy-pasted)
    """
    return secrets.token_urlsafe(nbytes=256)


def generate_user_code() -> str:
    """
    Returns an easy-to-type code like "WDJB-MJHT" with 20^8 entropy.
    Following the guidelines for user codes suggested here:
    https://tools.ietf.org/html/rfc8628#section-6.1
    """
    consonants = list(set(string.ascii_uppercase) - set("AEIOUY"))
    part1 = "".join(secrets.choice(consonants) for _ in range(4))
    part2 = "".join(secrets.choice(consonants) for _ in range(4))
    return f"{part1}-{part2}"


@router.post("/login")
def authorization_request(
    request: Request,
    db: Session = Depends(get_session),
    client_id: UUID = Form(...),
    scope: str = Form(""),
) -> Dict[str, Any]:
    """
    The device client initiates the authorization flow
    https://tools.ietf.org/html/rfc8628
    """
    logger.debug(f"Creating device code for device client login attempt {client_id}")
    conf = settings.auth.oauth_provider
    assert conf is not None
    device_code = generate_device_code()
    user_code = generate_user_code()
    verification_uri = f"{conf.redirect_scheme}://{request.url.netloc}" + VERIFICATION_PATH
    verification_uri_complete = verification_uri + f"?user_code={user_code}"
    expires_in = conf.device_code_lifetime.total_seconds()
    interval = conf.device_poll_interval.total_seconds()

    response = {
        "device_code": device_code,
        "user_code": user_code,
        "verification_uri": verification_uri,
        "verification_uri_complete": verification_uri_complete,
        "expires_in": expires_in,
        "interval": interval,
    }
    expiration_date = datetime.utcnow() + conf.device_code_lifetime
    users.create_device_code_attempt(
        db=db,
        client_id=client_id,
        expiration=expiration_date,
        device_code=device_code,
        user_code=user_code,
        scope=scope,
    )
    logger.debug(f"Device login flow with user-code {user_code} will expire at {expiration_date}")
    users.cleanup_device_code_attempts(db)
    db.commit()

    return response


@router.post("/token")
def access_token_request(
    db: Session = Depends(get_session),
    grant_type: str = Form(...),
    device_code: str = Form(...),
    client_id: UUID = Form(...),
) -> Dict[str, Any]:
    """
    The device client polls this endpoint until the user has approved the grant.
    https://tools.ietf.org/html/rfc8628#section-3.4
    """
    if grant_type != GRANT_TYPE:
        raise HTTPException(status_code=400, detail="unsupported_grant_type")

    try:
        logger.debug(f"Look up device_code attempt for device client {client_id}")
        device_code_attempt = users.get_device_code_attempt_by_client(db, client_id, device_code)
    except exc.NoResultFound:
        # Can't find that device code
        raise HTTPException(status_code=400, detail="invalid_client")

    if device_code_attempt.user_denied:
        # The user decided NOT to give Balsam their identity
        raise HTTPException(status_code=403, detail="access_denied")

    if datetime.utcnow() >= device_code_attempt.expiration:
        # This Device code login attempt is too old
        raise HTTPException(status_code=400, detail="expired_token")

    if device_code_attempt.user_id is None:
        # We haven't associated this device code with an ALCF user yet.
        # Client should wait a few seconds and try again.
        logger.debug(f"authorization pending for device code attempt {client_id}")
        raise HTTPException(status_code=400, detail="authorization_pending")

    user = UserOut(id=device_code_attempt.user.id, username=device_code_attempt.user.username)
    logger.debug(f"device code authorization success! associating {client_id} with username {user.username}")
    token, expiry = create_access_token(user)
    users.delete_device_code_attempt(db, device_code)
    db.commit()
    return {"access_token": token, "token_type": "bearer", "expiration": expiry}



import os
from typing import List, Optional, Union

from .compute_node import ComputeNode


class CoriHaswellNode(ComputeNode):

    cpu_ids = list(range(32))
    gpu_ids: List[Union[int, str]] = []

    @classmethod
    def get_job_nodelist(cls) -> List["CoriHaswellNode"]:
        """
        Get all compute nodes allocated in the current job context
        """
        nodelist_str = os.environ["SLURM_NODELIST"]
        # string like: nid[02529,02878,03047,03290,03331,03813,11847-11848]
        # or like: nid0[3038-3039,8241-8246]
        # or like: nid00[858-861]
        # or like: nid000[10-11]
        # or like: nid00558

        # first remove 'nid', '[' and ']'
        nodelist_str = nodelist_str.replace("nid", "").replace("[", "").replace("]", "")
        # now have something like '02529,02878,03047,03290,03331,03813,11847-11848'
        # split by comma
        node_ranges_str = nodelist_str.split(",")
        node_ids = []
        lo: Union[str, int]
        hi: Union[int, List[str]]
        for node_range_str in node_ranges_str:
            lo, *hi = node_range_str.split("-")
            lo = int(lo)
            if hi:
                hi = int(hi[0])
                node_ids.extend(list(range(lo, hi + 1)))
            else:
                node_ids.append(lo)

        return [cls(node_id, f"nid{node_id:05d}") for node_id in node_ids]

    @staticmethod
    def get_scheduler_id() -> Optional[int]:
        id = os.environ.get("SLURM_JOB_ID")
        if id is not None:
            return int(id)
        return None


if __name__ == "__main__":
    if "SLURM_NODELIST" not in os.environ:
        os.environ["SLURM_NODELIST"] = "nid0[3038-3039,8241-8246]"
    print([str(x) for x in CoriHaswellNode.get_job_nodelist()])



from typing import Any, Dict, List

from fastapi import APIRouter, Depends
from sqlalchemy import orm

from balsam import schemas
from balsam.server import settings
from balsam.server.models import TransferItem, crud, get_session
from balsam.server.pubsub import pubsub
from balsam.server.utils import Paginator

from .filters import TransferItemQuery

router = APIRouter()
auth = settings.auth.get_auth_method()


@router.get("/", response_model=schemas.PaginatedTransferItemOut)
def list(
    db: orm.Session = Depends(get_session),
    user: schemas.UserOut = Depends(auth),
    paginator: Paginator[TransferItem] = Depends(Paginator),
    q: TransferItemQuery = Depends(TransferItemQuery),
) -> Dict[str, Any]:
    """List data transfers associated with the user's Jobs."""
    count, transfers = crud.transfers.fetch(db, owner=user, paginator=paginator, filterset=q)
    return {"count": count, "results": transfers}


@router.get("/{transfer_id}", response_model=schemas.TransferItemOut)
def read(
    transfer_id: int, db: orm.Session = Depends(get_session), user: schemas.UserOut = Depends(auth)
) -> TransferItem:
    """Fetch a data transfer item by id."""
    count, transfers = crud.transfers.fetch(db, owner=user, transfer_id=transfer_id)
    assert isinstance(transfers, List)
    item: TransferItem = transfers[0]
    return item


@router.put("/{transfer_id}", response_model=schemas.TransferItemOut)
def update(
    transfer_id: int,
    data: schemas.TransferItemUpdate,
    db: orm.Session = Depends(get_session),
    user: schemas.UserOut = Depends(auth),
) -> schemas.TransferItemOut:
    """Update a transfer item by id."""
    updated_transfer, updated_job, log_event = crud.transfers.update(
        db, owner=user, transfer_id=transfer_id, data=data
    )
    result = schemas.TransferItemOut.from_orm(updated_transfer)
    db.commit()
    pubsub.publish(user.id, "update", "transfer", result)
    if updated_job:
        pubsub.publish(user.id, "update", "job", schemas.JobOut.from_orm(updated_job))
        pubsub.publish(user.id, "create", "log-event", schemas.LogEventOut.from_orm(log_event))
    return result


@router.patch("/", response_model=List[schemas.TransferItemOut])
def bulk_update(
    transfers: List[schemas.TransferItemBulkUpdate],
    db: orm.Session = Depends(get_session),
    user: schemas.UserOut = Depends(auth),
) -> List[schemas.TransferItemOut]:
    """Update a list of transfer items."""
    updated_transfers, updated_jobs, log_events = crud.transfers.bulk_update(db, owner=user, update_list=transfers)
    result_transfers = [schemas.TransferItemOut.from_orm(t) for t in updated_transfers]
    result_jobs = [schemas.JobOut.from_orm(j) for j in updated_jobs]
    result_events = [schemas.LogEventOut.from_orm(e) for e in log_events]
    db.commit()

    pubsub.publish(user.id, "bulk-update", "transfer", result_transfers)
    if result_jobs or result_events:
        pubsub.publish(user.id, "bulk-update", "job", result_jobs)
        pubsub.publish(user.id, "bulk-create", "event", result_events)
    return result_transfers



import logging
from typing import TYPE_CHECKING, Any

from balsam.util import Process, SigHandler

if TYPE_CHECKING:
    from balsam.client import RESTClient

logger = logging.getLogger(__name__)


class BalsamService(Process):
    def __init__(self, client: "RESTClient", *args: Any, service_period: float = 1.0, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.client = client
        self.service_period = service_period

    def _run(self, *args: Any, **kwargs: Any) -> None:
        self.sig_handler = SigHandler()
        self.client.close_session()

        while not self.sig_handler.wait_until_exit(timeout=self.service_period):
            self.run_cycle()
        logger.info(f"Signal: {self.__class__.__name__} Process cleaning up")
        self.cleanup()
        logger.info(f"{self.__class__.__name__} Process exit")

    def run_cycle(self) -> None:
        pass

    def cleanup(self) -> None:
        pass



import logging
from datetime import datetime, timedelta
from typing import List, Optional, cast
from uuid import UUID

from sqlalchemy.orm import Session
from sqlalchemy.sql import exists

from balsam.schemas import UserOut
from balsam.server.auth.password_utils import get_hash
from balsam.server.models.tables import AuthorizationState, DeviceCodeAttempt, User

logger = logging.getLogger(__name__)


def get_user_by_username(db: Session, username: str) -> User:
    return db.query(User).filter(User.username == username).one()


def user_exists(db: Session, username: str) -> bool:
    return bool(db.query(exists().where(User.username == username)).scalar())  # type: ignore


def create_user(db: Session, username: str, password: Optional[str]) -> UserOut:
    if password:
        hashed = get_hash(password)
        new_user = User(username=username, hashed_password=hashed)
    else:
        new_user = User(username=username)
    db.add(new_user)
    db.flush()
    return UserOut(id=new_user.id, username=new_user.username)


def list_users(db: Session) -> List[User]:
    return list(db.query(User))


def create_device_code_attempt(
    db: Session,
    client_id: UUID,
    expiration: datetime,
    device_code: str,
    user_code: str,
    scope: str,
) -> None:
    attempt = DeviceCodeAttempt(
        client_id=cast(str, client_id),
        expiration=expiration,
        device_code=device_code,
        user_code=user_code,
        scope=scope,
    )
    db.add(attempt)
    db.flush()


def cleanup_device_code_attempts(db: Session) -> None:
    cleanup_before = datetime.utcnow() - timedelta(minutes=30)
    qs = db.query(DeviceCodeAttempt)
    to_clean = qs.filter((DeviceCodeAttempt.expiration <= cleanup_before) | DeviceCodeAttempt.user_denied)
    logger.debug(f"Deleting {to_clean.count()} expired/denied device-code attempts")
    to_clean.delete(synchronize_session=False)
    db.flush()


def get_device_code_attempt_by_client(db: Session, client_id: UUID, device_code: str) -> DeviceCodeAttempt:
    qs = db.query(DeviceCodeAttempt).filter(
        DeviceCodeAttempt.client_id == client_id,
        DeviceCodeAttempt.device_code == device_code,
    )
    return qs.one()


def get_device_code_attempt_by_user(db: Session, user_code: str) -> DeviceCodeAttempt:
    qs = db.query(DeviceCodeAttempt).filter(DeviceCodeAttempt.user_code == user_code)
    return qs.one()


def authorize_device_code_attempt(db: Session, user_code: str, user: UserOut) -> None:
    device_code_attempt = get_device_code_attempt_by_user(db, user_code)
    device_code_attempt.user_id = user.id
    db.flush()


def deny_device_code_attempt(db: Session, user_code: str) -> None:
    device_code_attempt = get_device_code_attempt_by_user(db, user_code)
    device_code_attempt.user_denied = True
    db.flush()


def delete_device_code_attempt(db: Session, device_code: str) -> None:
    qs = db.query(DeviceCodeAttempt).filter(DeviceCodeAttempt.device_code == device_code)
    qs.one()
    qs.delete(synchronize_session=False)
    db.flush()


def add_auth_state(db: Session, state: str) -> None:
    auth_state = AuthorizationState(state=state)
    db.add(auth_state)
    db.flush()


def verify_auth_state(db: Session, state: str) -> None:
    """
    Raises exception if there is no matching state in DB
    """
    qs = db.query(AuthorizationState).filter(AuthorizationState.state == state)
    qs.one()
    qs.delete(synchronize_session=False)
    db.flush()



from datetime import datetime
from typing import Any, Dict

from fastapi import APIRouter, Depends, status
from sqlalchemy import orm

from balsam import schemas
from balsam.server import settings
from balsam.server.models import Site, crud, get_session
from balsam.server.pubsub import pubsub
from balsam.server.utils import Paginator

from .filters import SiteQuery

router = APIRouter()
auth = settings.auth.get_auth_method()


@router.get("/", response_model=schemas.PaginatedSitesOut)
def list(
    q: SiteQuery = Depends(SiteQuery),
    db: orm.Session = Depends(get_session),
    user: schemas.UserOut = Depends(auth),
    paginator: Paginator[Site] = Depends(Paginator),
) -> Dict[str, Any]:
    """List Sites belonging to the user."""
    count, sites = crud.sites.fetch(
        db,
        owner=user,
        paginator=paginator,
        filterset=q,
    )
    return {"count": count, "results": sites}


@router.get("/{site_id}", response_model=schemas.SiteOut)
def read(site_id: int, db: orm.Session = Depends(get_session), user: schemas.UserOut = Depends(auth)) -> Site:
    """Fetch a Sites by id."""
    _, site = crud.sites.fetch(db, owner=user, site_id=site_id)
    assert isinstance(site, Site)
    return site


@router.post("/", response_model=schemas.SiteOut, status_code=status.HTTP_201_CREATED)
def create(
    site: schemas.SiteCreate,
    db: orm.Session = Depends(get_session),
    user: schemas.UserOut = Depends(auth),
) -> schemas.SiteOut:
    """Create a new Balsam Site."""
    new_site = crud.sites.create(db, owner=user, site=site)
    result = schemas.SiteOut.from_orm(new_site)
    db.commit()
    pubsub.publish(user.id, "create", "site", result)
    return result


@router.put("/{site_id}", response_model=schemas.SiteOut)
def update(
    site_id: int,
    site: schemas.SiteUpdate,
    db: orm.Session = Depends(get_session),
    user: schemas.UserOut = Depends(auth),
) -> Site:
    """Update a Balsam Site by id."""
    data = site.dict(exclude_unset=True)
    data["last_refresh"] = datetime.utcnow()
    updated_site = crud.sites.update(db, owner=user, site_id=site_id, update_data=data)
    data["id"] = site_id
    db.commit()
    pubsub.publish(user.id, "update", "site", data)
    return updated_site


@router.delete("/{site_id}", status_code=status.HTTP_204_NO_CONTENT)
def delete(site_id: int, db: orm.Session = Depends(get_session), user: schemas.UserOut = Depends(auth)) -> None:
    """Delete a Balsam Site by id. All associated Apps, Jobs, and BatchJobs are destroyed."""
    crud.sites.delete(db, owner=user, site_id=site_id)
    db.commit()
    pubsub.publish(user.id, "delete", "site", {"id": site_id})



import subprocess
import sys
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Type

import jinja2
from pydantic import BaseModel, PyObject
from pydantic.fields import ModelField

from balsam._api.model import BalsamModel

FieldDict = Dict[str, Any]

env = jinja2.Environment(trim_blocks=True, lstrip_blocks=True)
header_template = env.from_string(
    """
# This file was auto-generated via {{ generator_name }}
# [git rev {{git_ref}}]
# Do *not* make changes to the API by changing this file!

{% for imp_statement in import_modules -%}
{{ imp_statement }}
{% endfor %}

    """
)
master_template = env.from_string(
    """
class {{model_name}}({{model_base}}):
    _create_model_cls = {{_create_model_cls}}
    _update_model_cls = {{_update_model_cls}}
    _read_model_cls = {{_read_model_cls}}
    objects: "{{manager_name}}"

    {% for field in model_fields.values() %}
    {% if field.allowed_none and 'Optional' not in field.annotation %}
    {{field.name}} = Field[Optional[{{field.annotation}}]]()
    {% else %}
    {{field.name}} = Field[{{field.annotation}}]()
    {% endif %}
    {% endfor %}

    {% if model_create_kwargs %}
    def __init__(
        self,
        {% for line in model_create_kwargs %}
        {{line}},
        {% endfor %}
        **kwargs: Any,
    ) -> None:
        '''
        Construct a new {{model_name}} object.  You must eventually call the save() method or
        pass a {{model_name}} list into {{model_name}}.objects.bulk_create().

        {% for line in model_create_help %}
        {{line}}
        {% endfor %}
        '''
        _kwargs = {k: v for k,v in locals().items() if k not in ["self", "__class__"] and v is not None}
        _kwargs.update(kwargs)
        return super().__init__(**_kwargs)
    {% endif %}

class {{query_name}}(Query[{{model_name}}]):
    {% if not model_filter_kwargs and not order_by_type %}
    pass
    {% endif %}
    {% if model_filter_kwargs %}
    def get(
        self,
        {% for line in model_filter_kwargs %}
        {{line}},
        {% endfor %}
    ) -> {{model_name}}:
        '''
        Retrieve exactly one {{model_name}}. Raises {{model_name}}.DoesNotExist
        if no items were found, or {{model_name}}.MultipleObjectsReturned if
        more than one item matched the query.

        {% for line in model_filter_help %}
        {{line}}
        {% endfor %}
        '''
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return self._get(**kwargs)

    def filter(
        self,
        {% for line in model_filter_kwargs %}
        {{line}},
        {% endfor %}
    ) -> "{{query_name}}":
        '''
        Retrieve exactly one {{model_name}}. Raises {{model_name}}.DoesNotExist
        if no items were found, or {{model_name}}.MultipleObjectsReturned if
        more than one item matched the query.

        {% for line in model_filter_help %}
        {{line}}
        {% endfor %}
        '''
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return self._filter(**kwargs)
    {% endif %}

    {% if model_update_kwargs %}
    def update(
        self,
        {% for line in model_update_kwargs %}
        {{line}},
        {% endfor %}
    ) -> List[{{model_name}}]:
        '''
        Updates all items selected by this query with the given values.

        {% for line in model_update_help %}
        {{line}}
        {% endfor %}
        '''
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return self._update(**kwargs)
    {% endif %}

    {% if order_by_type %}
    def order_by(self, field: Optional[{{order_by_type}}]) -> "{{query_name}}":
        '''
        Order the returned items by this field.
        '''
        return self._order_by(field)
    {% endif %}

class {{manager_name}}({{manager_base}}):
    _api_path = "{{manager_url}}"
    _model_class = {{model_name}}
    _query_class = {{query_name}}
    _bulk_create_enabled = {{_bulk_create_enabled}}
    _bulk_update_enabled = {{_bulk_update_enabled}}
    _bulk_delete_enabled = {{_bulk_delete_enabled}}
    _paginated_list_response = {{_paginated_list_response}}

    {% if model_create_kwargs %}
    def create(
        self,
        {% for line in model_create_kwargs %}
        {{line}},
        {% endfor %}
    ) -> {{model_name}}:
        '''
        Create a new {{model_name}} object and save it to the API in one step.

        {% for line in model_create_help %}
        {{line}}
        {% endfor %}
        '''
        kwargs = {k: v for k,v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return super()._create(**kwargs)
    {% endif %}

    def all(self) -> "{{query_name}}":
        '''
        Returns a Query for all {{model_name}} items.
        '''
        return self._query_class(manager=self)

    {% if model_filter_kwargs %}
    def get(
        self,
        {% for line in model_filter_kwargs %}
        {{line}},
        {% endfor %}
    ) -> {{model_name}}:
        '''
        Retrieve exactly one {{model_name}}. Raises {{model_name}}.DoesNotExist
        if no items were found, or {{model_name}}.MultipleObjectsReturned if
        more than one item matched the query.

        {% for line in model_filter_help %}
        {{line}}
        {% endfor %}
        '''
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return {{query_name}}(manager=self).get(**kwargs)

    def filter(
        self,
        {% for line in model_filter_kwargs %}
        {{line}},
        {% endfor %}
    ) -> "{{query_name}}":
        '''
        Returns a {{model_name}} Query returning items matching the filter criteria.

        {% for line in model_filter_help %}
        {{line}}
        {% endfor %}
        '''
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return {{query_name}}(manager=self).filter(**kwargs)
    {% endif %}

"""
)


def get_git_hash() -> str:
    return subprocess.check_output("git rev-parse --short HEAD", shell=True, encoding="utf-8").strip()


def field_to_dict(field: ModelField, schema: Type[BaseModel]) -> FieldDict:
    field_default = field.field_info.default
    if field_default == [] or field_default == {}:
        assert not field.required
        default_create = None
    elif not field.required:
        default_create = field_default
    else:
        assert field_default == ..., f"Expected required field {field} of {schema} to have default=Ellipsis"
        assert field.required
        default_create = ...
    if field.is_complex():
        annotation = None
        for cls in schema.mro():
            if field.name in cls.__annotations__:
                annotation = str(cls.__annotations__[field.name])
                break
        else:
            raise RuntimeError(f"Could not find annotation for {field.name} on {schema.__name__} MRO")
    else:
        annotation = qual_path(field.type_)

    assert annotation is not None
    annotation = annotation.replace("NoneType", "None")
    no_constructor = field.field_info.extra.get("no_constructor", False)
    no_descriptor = field.field_info.extra.get("no_descriptor", False)
    return {
        "name": field.name,
        "required": field.required,
        "description": getattr(field.field_info, "description", ""),
        "annotation": annotation,
        "schema_default": field_default,  # default attribute on the Pydantic ModelField
        "default_create": default_create,  # default value for __init__ kwargs
        "optional_create": default_create is None,  # whether to use Optional[] annotation in __init__
        "no_constructor": no_constructor,
        "no_descriptor": no_descriptor,
    }


def get_schema_fields(schema: Type[BaseModel]) -> Dict[str, FieldDict]:
    return {k: field_to_dict(v, schema) for k, v in schema.__fields__.items()}


def model_create_signature(create_fields: FieldDict) -> List[str]:
    fields = create_fields.copy()
    required_fields = dict((k, v) for k, v in fields.items() if v["required"])
    for k in required_fields:
        del fields[k]

    result = []
    for field in required_fields.values():
        result += [f'{field["name"]}: {field["annotation"]}']
    for field in fields.values():
        if field["optional_create"]:
            result += [f'{field["name"]}: Optional[{field["annotation"]}] = {repr(field["default_create"])}']
        else:
            result += [f'{field["name"]}: {field["annotation"]} = {repr(field["default_create"])}']
    return result


def filter_signature(filterset: object) -> List[str]:
    result = []
    for field in filterset.__dataclass_fields__.values():  # type: ignore
        if field.name == "ordering":
            continue
        try:
            field_type = qual_path(field.type)
        except AttributeError:
            field_type = str(field.type)
        if getattr(field.type, "_name", None) in ["List", "Set"]:
            inner_type = qual_path(field.type.__args__[0])
            annotation = f"Union[{field_type}, {inner_type}, None]"
        else:
            annotation = f"Optional[{field_type}]"
        kwarg = f"{field.name}: {annotation} = None"
        result.append(kwarg)
    return result


def update_signature(update_fields: FieldDict) -> List[str]:
    result = []
    for field in update_fields.values():
        result += [f'{field["name"]}: Optional[{field["annotation"]}] = None']
    return result


def order_by_typename(filterset: object) -> Optional[str]:
    if "ordering" not in filterset.__dataclass_fields__:  # type: ignore
        return None
    order_enum = filterset.__dataclass_fields__["ordering"].type  # type: ignore
    assert issubclass(order_enum, Enum)
    typename = qual_path(order_enum)
    return typename


def get_model_fields(model_base: Type[BalsamModel]) -> Tuple[FieldDict, FieldDict, FieldDict]:
    create_model = model_base._create_model_cls
    update_model = model_base._update_model_cls
    read_model = model_base._read_model_cls

    create_fields = get_schema_fields(create_model) if create_model is not None else {}
    update_fields = get_schema_fields(update_model) if update_model is not None else {}
    read_fields = get_schema_fields(read_model)
    return create_fields, update_fields, read_fields


def qual_path(obj: type) -> str:
    if obj is None:
        return "None"
    mod = str(obj.__module__)
    try:
        name = str(obj.__name__)
    except AttributeError:
        return str(obj)
    else:
        if mod == "builtins":
            return name
        return f"{mod}.{name}"


def make_help_text(fields: FieldDict) -> List[str]:
    result: List[str] = []
    if not fields:
        return result
    maxwidth = max(len(f) for f in fields.keys()) + 1
    for field_name, field in fields.items():
        if field_name == "ordering":
            continue
        try:
            descr = field["description"]
        except Exception:
            descr = field.default.description
        result.append(f"{(field_name+':').ljust(maxwidth)} {descr}")
    return result


def get_model_ctx(model_base: Type[BalsamModel], manager_base: type, filterset: type) -> Dict[str, Any]:
    base_name = model_base.__name__
    name = base_name[: base_name.find("Base")]
    manager_name = f"{name}Manager"
    query_name = f"{name}Query"
    base_name = qual_path(model_base)

    create_fields, update_fields, read_fields = get_model_fields(model_base)
    constructor_fields = {k: v for k, v in create_fields.items() if not v["no_constructor"]}
    create_kwargs = model_create_signature(constructor_fields) if create_fields else None
    update_kwargs = update_signature(update_fields) if update_fields else None
    filter_kwargs = filter_signature(filterset)
    order_by_type = order_by_typename(filterset)

    model_create_help = make_help_text(constructor_fields)
    model_update_help = make_help_text(update_fields)
    model_filter_help = make_help_text(filterset.__dataclass_fields__)  # type: ignore
    fields = {**create_fields, **update_fields, **read_fields}
    fields = {k: v for k, v in fields.items() if not v["no_descriptor"]}
    for field in fields:
        # A read-only field can be None if the model is creatable (e.g. not created yet, id is None)
        if field in read_fields and field not in create_fields and model_base._create_model_cls is not None:
            fields[field]["allowed_none"] = True
        # A write-only field can be None (e.g. state_timestamp not written yet)
        elif field in update_fields and field not in read_fields:
            fields[field]["allowed_none"] = True
        # A field that can actually be null in the API:
        elif field in read_fields and read_fields[field]["schema_default"] is None:
            fields[field]["allowed_none"] = True
        else:
            fields[field]["allowed_none"] = False

    return dict(
        model_name=name,
        model_base=base_name,
        _create_model_cls=qual_path(model_base.__dict__["_create_model_cls"]),
        _update_model_cls=qual_path(model_base.__dict__["_update_model_cls"]),
        _read_model_cls=qual_path(model_base.__dict__["_read_model_cls"]),
        manager_name=manager_name,
        query_name=query_name,
        model_fields=fields,
        model_create_kwargs=create_kwargs,
        manager_base=qual_path(manager_base),
        manager_url=manager_base._api_path,  # type: ignore
        _bulk_create_enabled=getattr(manager_base, "_bulk_create_enabled", False),
        _bulk_update_enabled=getattr(manager_base, "_bulk_update_enabled", False),
        _bulk_delete_enabled=getattr(manager_base, "_bulk_delete_enabled", False),
        _paginated_list_response=getattr(manager_base, "_paginated_list_response", True),
        model_update_kwargs=update_kwargs,
        model_filter_kwargs=filter_kwargs,
        order_by_type=order_by_type,
        model_create_help=model_create_help,
        model_filter_help=model_filter_help,
        model_update_help=model_update_help,
    )


class APIConf(BaseModel):
    model_base: PyObject
    manager_base: PyObject
    filterset: PyObject


def main() -> None:
    imports = [
        "import datetime",
        "import typing",
        "from typing import Optional, Any, List, Union",
        "import pathlib",
        "import uuid",
        "import pydantic",
        "import balsam._api.model",
        "import balsam._api.bases",
        "from balsam._api.query import Query",
        "from balsam._api.model import Field",
    ]
    header = header_template.render(
        generator_name=f"{sys.executable} {' '.join(sys.argv)}",
        git_ref=get_git_hash(),
        import_modules=imports,
    )
    print(header)

    models = [
        "Site",
        "App",
        "Job",
        "BatchJob",
        "Session",
        "TransferItem",
        "EventLog",
    ]
    for model in models:
        conf = APIConf(
            model_base=f"balsam._api.bases.{model}Base",
            manager_base=f"balsam._api.bases.{model}ManagerBase",
            filterset=f"balsam.server.routers.filters.{model}Query",
        )
        model_ctx = get_model_ctx(conf.model_base, conf.manager_base, conf.filterset)  # type: ignore
        result = master_template.render(**model_ctx)
        print(result)


if __name__ == "__main__":
    main()



import logging
from datetime import datetime
from itertools import accumulate
from typing import Dict, List, Tuple

from balsam._api.models import BatchJobQuery, EventLogQuery
from balsam.schemas import EventOrdering

logger = logging.getLogger(__name__)


def throughput_report(log_query: EventLogQuery, to_state: str = "JOB_FINISHED") -> Tuple[List[datetime], List[int]]:
    finished_logs = log_query.filter(to_state=to_state).order_by(EventOrdering.timestamp)
    times = list(log.timestamp for log in finished_logs)
    return times, list(range(1, len(times) + 1))


def utilization_report(log_query: EventLogQuery, node_weighting: bool = True) -> Tuple[List[datetime], List[float]]:
    job_events = []
    nodes_by_id: Dict[int, float] = {}
    for log in log_query.filter(to_state="RUNNING"):
        usage_count = log.data["num_nodes"] if node_weighting else 1.0
        nodes_by_id[log.job_id] = usage_count
        job_events.append((log.timestamp, usage_count))
    for log in log_query.filter(from_state="RUNNING"):
        job_events.append((log.timestamp, -1.0 * nodes_by_id.get(log.job_id, 1.0)))

    util_times, util_counts = zip(*sorted(job_events))
    return list(util_times), list(accumulate(util_counts))


def available_nodes(batch_job_query: BatchJobQuery) -> Tuple[List[datetime], List[int]]:
    running: List[Tuple[datetime, int]] = []
    for job in batch_job_query:
        if job.start_time is None or job.end_time is None:
            logger.warning(f"Skipping BatchJob {job.id}: missing start/end times")
            continue
        running.append((job.start_time, job.num_nodes))
        running.append((job.end_time, -1 * job.num_nodes))

    if not running:
        return [], []
    running_nodes_times, running_node_counts = zip(*sorted(running))
    return list(running_nodes_times), list(accumulate(running_node_counts))



import os

from .app_run import SubprocessAppRun


class SummitJsrun(SubprocessAppRun):
    """
    https://docs.olcf.ornl.gov/systems/summit_user_guide.html#running-jobs
    """

    def _build_cmdline(self) -> str:
        cpu_per_rs = self.get_cpus_per_rank()
        args = [
            "jsrun",
            "--nrs",  # number of Resource Sets == number of ranks
            self.get_num_ranks(),
            "--tasks_per_rs",  # ranks per resource set
            1,  # 1 rank per resource set (fixed assumption)
            "--rs_per_host",
            self._ranks_per_node,
            "--cpu_per_rs",
            cpu_per_rs,  # number of cores per resource set
            "--bind",
            "rs",  # Bind to all cores of resource set
            "--gpu_per_rs",
            self._gpus_per_rank,  # gpus per resource set
            self._cmdline,
        ]
        return " ".join(str(arg) for arg in args)

    def _set_envs(self) -> None:
        # Override because we do not want/need to set
        # CUDA_VISIBLE_DEVICES by default on Summit
        envs = os.environ.copy()
        envs.update(self._envs)
        envs["OMP_NUM_THREADS"] = str(self._threads_per_rank)
        self._envs = envs



import logging
import os
import subprocess
from pathlib import Path
from typing import List, Sequence, Tuple, Union
from uuid import UUID

from globus_sdk import TransferData  # type: ignore
from globus_sdk.exc import GlobusConnectionError, TransferAPIError  # type: ignore

from balsam.util.globus_auth import get_client  # type: ignore

from .transfer import TaskInfo, TransferInterface, TransferRetryableError, TransferSubmitError

logger = logging.getLogger(__name__)


PathLike = Union[str, Path]
SrcDestRecursive = Tuple[PathLike, PathLike, bool]


def submit_sdk(src_endpoint: UUID, dest_endpoint: UUID, batch: Sequence[SrcDestRecursive]) -> str:
    client = get_client()
    notify_opts = {
        "notify_on_succeeded": False,
        "notify_on_failed": False,
        "notify_on_inactive": False,
    }
    transfer_data = TransferData(
        client,
        src_endpoint,
        dest_endpoint,
        label="",
        sync_level="size",
        verify_checksum=False,
        preserve_timestamp=False,
        encrypt_data=True,
        submission_id=None,
        delete_destination_extra=False,
        deadline=None,
        skip_activation_check=False,
        **notify_opts,
    )
    for src, dest, recurse in batch:
        transfer_data.add_item(str(src), str(dest), recursive=recurse)
    try:
        res = client.submit_transfer(transfer_data)
    except TransferAPIError as exc:
        raise TransferSubmitError(str(exc))
    task_id = res.get("task_id", None)
    if task_id is None:
        raise TransferSubmitError(str(res))
    logger.info(f"Submitted Globus transfer task {task_id}")
    return "globus:" + str(task_id)


def submit_subproc(src_endpoint: UUID, dest_endpoint: UUID, batch: List[SrcDestRecursive]) -> str:
    env = os.environ.copy()
    env["LC_ALL"] = "C.UTF-8"
    env["LANG"] = "C.UTF-8"
    batch_str = "\n".join(f"{src} {dest} {'-r' if recurse else ''}" for src, dest, recurse in batch)
    proc = subprocess.run(
        [
            "globus",
            "transfer",
            "--encrypt",
            "--batch",
            str(src_endpoint),
            str(dest_endpoint),
        ],
        input=batch_str,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        encoding="utf-8",
        env=env,
        check=True,
    )
    for line in proc.stdout.split("\n"):
        if "Task ID" in line:
            return "globus:" + line.split()[-1]
    raise TransferSubmitError("Could not parse Task ID from submission")


class GlobusTransferInterface(TransferInterface):
    def __init__(self, endpoint_id: UUID):
        self.endpoint_id: UUID = UUID(str(endpoint_id))

    @staticmethod
    def _state_map(status: str) -> str:
        return {
            "active": "active",
            "inactive": "inactive",
            "succeeded": "done",
            "failed": "error",
        }[status.strip().lower()]

    def submit_task(
        self,
        remote_loc: str,
        direction: str,
        transfer_paths: Sequence[Tuple[PathLike, PathLike, bool]],
    ) -> str:
        """Submit Transfer Task via Globus CLI"""
        if direction == "in":
            src_endpoint, dest_endpoint = UUID(str(remote_loc)), self.endpoint_id
        elif direction == "out":
            src_endpoint, dest_endpoint = self.endpoint_id, UUID(str(remote_loc))
        else:
            raise ValueError("direction must be in or out")
        try:
            task_id = submit_sdk(src_endpoint, dest_endpoint, transfer_paths)
        except GlobusConnectionError as exc:
            raise TransferRetryableError(f"GlobusConnectionError in Transfer task submission: {exc}") from exc
        return str(task_id)

    @staticmethod
    def _poll_tasks(task_ids: Sequence[str]) -> List[TaskInfo]:
        client = get_client()
        filter_values = {"task_id": ",".join(map(str, task_ids))}
        filter_str = "/".join(f"{k}:{v}" for k, v in filter_values.items())
        try:
            task_list = list(client.task_list(num_results=None, filter=filter_str))
        except GlobusConnectionError as exc:
            raise TransferRetryableError(f"GlobusConnectionError in client.task_list: {exc}")
        result = []
        for d in task_list:
            state = GlobusTransferInterface._state_map(d["status"])
            info = {}
            if d.get("fatal_error"):
                info["error"] = d["fatal_error"]
            task_id = "globus:" + d["task_id"]
            result.append(TaskInfo(task_id=task_id, state=state, info=info))
        return result


TransferInterface._registry["globus"] = GlobusTransferInterface



import click
import yaml

from balsam.client.encoders import jsonable_encoder
from balsam.schemas import UserOut
from balsam.server.auth.token import create_access_token
from balsam.server.models import get_session
from balsam.server.models.crud.users import create_user, get_user_by_username


@click.group()
def cli() -> None:
    pass


@cli.command()
@click.argument("username")
def new_user(username: str) -> None:
    db = next(get_session())
    userout = create_user(db, username, password=None)
    db.commit()
    print("Created user", userout)


@cli.command()
@click.argument("username")
def make_token(username: str) -> None:
    db = next(get_session())
    from_db = get_user_by_username(db, username)
    user_out = UserOut(id=from_db.id, username=from_db.username)
    jwt, expiry = create_access_token(user_out)
    s = yaml.dump(
        jsonable_encoder(
            {
                "api_root": "https://balsam-dev.alcf.anl.gov",
                "username": username,
                "client_class": "balsam.client.BasicAuthRequestsClient",
                "token": jwt,
                "token_expiry": expiry,
                "connect_timeout": 6.2,
                "read_timeout": 120.0,
                "retry_count": 10,
            }
        )
    )
    print(s)


if __name__ == "__main__":
    cli()



import json
import logging
from typing import Any, Optional

import aredis  # type: ignore
import redis
from fastapi.encoders import jsonable_encoder

from balsam.server import settings

logger = logging.getLogger(__name__)


class _PubSub:
    """
    Sync Publisher and Async Subscriber (for Websockets)
    """

    def __init__(self) -> None:
        self._redis: Optional[redis.Redis[str]] = None
        self._async_redis = None
        self._has_warned = False

    @property
    def r(self) -> "redis.Redis[str]":
        if self._redis is None:
            self._redis = redis.Redis(**settings.redis_params)
        return self._redis

    @property
    def async_r(self) -> aredis.StrictRedis:
        if self._async_redis is None:
            self._async_redis = aredis.StrictRedis(**settings.redis_params)
        return self._async_redis

    @staticmethod
    def get_topic(user_id: int) -> str:
        return f"user-{user_id}"

    async def subscribe(self, user_id: int) -> "aredis.PubSub":
        p = self.async_r.pubsub(ignore_subscribe_messages=True)
        await p.subscribe(self.get_topic(user_id))
        return p

    def publish(self, user_id: int, action: str, type: str, data: Any) -> None:
        msg = {"action": action, "type": type, "data": data}
        try:
            self.r.publish(self.get_topic(user_id), json.dumps(jsonable_encoder(msg)))
        except redis.exceptions.ConnectionError as e:
            if not self._has_warned:
                logger.warning(f"Redis connection failed!\n{e}\n Proceeding without Redis pub/sub.")
                self._has_warned = True


pubsub = _PubSub()



from balsam.api import ApplicationDefinition


class BashRunner(ApplicationDefinition):
    """
    Parsl Bash Runner. Place this file in your site directory under apps/ directory.
    """

    site = 0
    command_template = "/bin/bash -c {{ command }}"

    def preprocess(self):
        print("COMMAND:", self.command_template)
        self.job.state = "PREPROCESSED"

    def postprocess(self):
        from balsam.api import site_config

        workdir = self.job.resolve_workdir(site_config.data_path)
        stdout = workdir.joinpath("job.out").read_text().strip()
        self.job.data = {"result": stdout, "type": "bash"}
        self.job.state = "POSTPROCESSED"

    def shell_preamble(self):
        pass

    def handle_timeout(self):
        self.job.state = "RESTART_READY"

    def handle_error(self):
        self.job.state = "FAILED"


class AppRunner(ApplicationDefinition):
    """
    Parsl App Runner. Place this file in your site directory under apps/ directory.
    """

    site = 0
    environment_variables = {}
    command_template = "singularity exec --bind {{ workdir }}:/work --bind .:/app {{ image }} {{ command }}"
    parameters = {}
    transfers = {}

    def preprocess(self):
        print("COMMAND:", self.command_template)
        self.job.state = "PREPROCESSED"

    def postprocess(self):
        import json

        from balsam.api import site_config

        workdir = self.job.resolve_workdir(site_config.data_path)
        print("WORKDIR: ", workdir)
        stdout = workdir.joinpath("job.out").read_text().strip()
        metadata = workdir.joinpath("job.metadata").read_text().strip()
        print("STDOUT2: ", stdout)

        print("METADATA: ", metadata)
        try:
            metadata = json.loads(metadata)
            self.job.data = metadata
        except:  # noqa
            import traceback

            print(traceback.format_exc())
            self.job.data = {"result": stdout, "type": "bash"}

        self.job.state = "POSTPROCESSED"

    def shell_preamble(self):
        pass

    def handle_timeout(self):
        self.job.state = "RESTART_READY"

    def handle_error(self):
        self.job.state = "FAILED"



import getpass

import click

from balsam.client import NotAuthenticatedError, RequestsClient, urls
from balsam.config import ClientSettings


def is_auth() -> bool:
    """
    Returns True only if an access token is stored and working.
    """
    try:
        settings = ClientSettings.load_from_file()
    except NotAuthenticatedError:
        return False

    client = settings.build_client()
    return client._authenticated


@click.command()
@click.option("-u", "--url", default="https://balsam-dev.alcf.anl.gov", help="Balsam server address")
@click.option("-f", "--force", is_flag=True, default=False, help="Force redo, even if logged in")
def login(url: str, force: bool) -> None:
    """
    Set client information and authenticate to server
    """
    if not force and is_auth():
        click.echo("You already have a good login stored.")
        click.echo("To force login, try again with the --force option.")
        return

    url = url.strip()
    if not url.startswith("http"):
        url = "https://" + url
    click.echo(f"Logging into {url}")

    client = RequestsClient.discover_supported_client(url)
    settings = ClientSettings(api_root=url, client_class=type(client))

    update_fields = client.interactive_login()
    if update_fields:
        data = settings.dict()
        data.update(update_fields)
        updated_settings = ClientSettings(**data)
        updated_settings.save_to_file()


@click.command()
@click.option("-a", "--address", prompt="Balsam server address", help="Balsam server address")
@click.option("-u", "--username", prompt="Balsam username", help="Balsam username")
def register(address: str, username: str) -> None:
    """
    Register a new user account with Balsam server
    """
    settings = ClientSettings(api_root=address)
    client = settings.build_client()
    password = getpass.getpass("Password:")
    conf_password = getpass.getpass("Confirm Password:")
    if password != conf_password:
        raise click.BadParameter("Passwords must match")

    resp = client.post(urls.PASSWORD_REGISTER, username=username, password=password, authenticating=True)
    click.echo(f"Registration success! {resp}")



import logging
import time
from datetime import datetime, timedelta
from typing import Any, Dict, Optional
from uuid import UUID, uuid4

import click
import dateutil
from requests import HTTPError

from balsam.util import Spinner

from . import urls
from .requests_client import RequestsClient
from .rest_base_client import AuthError

logger = logging.getLogger(__name__)


class OAuthRequestsClient(RequestsClient):
    def __init__(
        self,
        api_root: str,
        token: Optional[str] = None,
        token_expiry: Optional[datetime] = None,
        connect_timeout: float = 6.2,
        read_timeout: float = 120.0,
        retry_count: int = 10,
    ) -> None:
        super().__init__(
            api_root,
            connect_timeout=connect_timeout,
            read_timeout=read_timeout,
            retry_count=retry_count,
        )
        self.token = token
        self.token_expiry = token_expiry
        self.expires_in = timedelta(hours=240)
        self.login_attempt_id: Optional[UUID] = None
        if self.token is not None:
            self.session.auth = None
            self.session.headers["Authorization"] = f"Bearer {self.token}"
            self._authenticated = True
        if self.token_expiry:
            self.expires_in = self.token_expiry - datetime.utcnow()
            if self.expires_in.total_seconds() <= 1:
                self._authenticated = False
                logger.warning("Auth Token is expired; please refresh with `balsam login`")
            elif self.expires_in < timedelta(hours=24):
                logger.warning(f"Auth Token will expire in {self.expires_in}; please refresh with `balsam login`")

    def refresh_auth(self) -> None:
        self.session.auth = None
        self.session.headers["Authorization"] = f"Bearer {self.token}"
        self._authenticated = True

    def prompt_user(self, auth_uri_complete: str) -> None:
        click.echo(f"Logging into Balsam API at {self.api_root}")
        click.echo(f"To proceed, please navigate to: {auth_uri_complete}")
        click.echo("Authenticate with your credentials then come back here!")

    def poll_for_token(self, device_code: str, user_code: str, expires_in: float, interval: float) -> None:
        data = {
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            "device_code": device_code,
            "client_id": self.login_attempt_id,
        }
        num_attempts = int(expires_in / interval)
        for _ in range(num_attempts):
            time.sleep(interval)
            try:
                resp = self.request(urls.DEVICE_TOKEN, "POST", data=data, authenticating=True)
                break
            except HTTPError as exc:
                if exc.response.status_code == 400 and "authorization_pending" in exc.response.text:
                    continue
                raise
        else:
            raise AuthError(f"Login attempt timed out after {expires_in} seconds. Please try again.")

        if isinstance(resp, dict) and "access_token" in resp:
            self.token = resp["access_token"]
            self.token_expiry = dateutil.parser.parse(resp["expiration"])
        else:
            raise AuthError(f"Could not authenticate: {resp}")

    def interactive_login(self) -> Dict[str, Any]:
        """Initiate interactive login flow"""
        self.login_attempt_id = uuid4()
        resp = self.request(urls.DEVICE_LOGIN, "POST", data={"client_id": self.login_attempt_id}, authenticating=True)
        assert isinstance(resp, dict)
        if not (resp.get("device_code") and resp.get("user_code")):
            raise AuthError(f"Did not get expected device authorization response: {resp}")

        resp.pop("verification_uri")
        auth_uri_comp = resp.pop("verification_uri_complete")
        self.prompt_user(auth_uri_comp)
        with Spinner("Waiting for user log in..."):
            self.poll_for_token(**resp)
        self.refresh_auth()
        click.echo("Logged In")
        return {"token": self.token, "token_expiry": self.token_expiry}


RequestsClient._client_classes["oauth_device"] = OAuthRequestsClient



import logging
import resource
import time
from typing import Awaitable, Callable

from fastapi import APIRouter, FastAPI, Request, Response
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.routing import Match, Mount

logger = logging.getLogger(__name__)


def _get_cpu_time() -> float:
    """
    Generates the cpu time to report. Adds the user and system time, following the implementation from timing-asgi
    """
    resources = resource.getrusage(resource.RUSAGE_SELF)
    # add up user time (ru_utime) and system time (ru_stime)
    return resources[0] + resources[1]


class TimingMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: FastAPI, router: APIRouter):
        super().__init__(app)
        self.router = router

    def get_route_name(self, request: Request) -> str:
        scope = request.scope
        route = None
        for r in self.router.routes:
            if r.matches(scope)[0] == Match.FULL:
                route = r
                break
        if hasattr(route, "endpoint") and hasattr(route, "name"):
            name = f"{route.endpoint.__module__}.{route.name}"  # type: ignore
        elif isinstance(route, Mount):
            name = f"{type(route.app).__name__}<{route.name!r}>"
        else:
            name = str(f"<Path: {scope['path']}>")
        return name

    async def dispatch(self, request: Request, call_next: Callable[[Request], Awaitable[Response]]) -> Response:
        start_wall, start_cpu = time.perf_counter(), _get_cpu_time()
        response = await call_next(request)
        wall, cpu = time.perf_counter() - start_wall, _get_cpu_time() - start_cpu
        timing_info = f"TIMER: {self.get_route_name(request)} Wall: {wall:.6f} CPU: {cpu:.6f}"
        logger.info(timing_info)
        return response



from fastapi import HTTPException, status

from .conf import Settings

__version__ = "0.1"


class ValidationError(HTTPException):
    def __init__(self, detail: str) -> None:
        super().__init__(status_code=status.HTTP_400_BAD_REQUEST, detail=detail)


settings = Settings()
__all__ = ["settings", "Settings", "ValidationError"]



from .globus_transfer import GlobusTransferInterface
from .transfer import TaskInfo, TransferInterface, TransferRetryableError, TransferSubmitError

__all__ = [
    "TransferInterface",
    "GlobusTransferInterface",
    "TransferSubmitError",
    "TransferRetryableError",
    "TaskInfo",
]



from datetime import datetime
from enum import Enum
from typing import Any, Dict, List

from pydantic import BaseModel, Field


class EventOrdering(str, Enum):
    timestamp = "timestamp"
    timestamp_desc = "-timestamp"


class LogEventOut(BaseModel):
    id: int = Field(...)
    job_id: int = Field(...)
    timestamp: datetime = Field(...)
    # Do not validate on the way out:
    from_state: str = Field(...)
    to_state: str = Field(...)
    data: Dict[str, Any] = Field(...)

    class Config:
        orm_mode = True


class PaginatedLogEventOut(BaseModel):
    count: int
    results: List[LogEventOut]



import logging
import socket
from typing import List, Optional, Union

import psutil  # type: ignore

from .compute_node import ComputeNode

logger = logging.getLogger(__name__)


class DefaultNode(ComputeNode):

    cpu_ids = list(range(psutil.cpu_count() or 4))
    gpu_ids: List[Union[int, str]] = []

    @classmethod
    def get_job_nodelist(cls) -> List["DefaultNode"]:
        hostname = socket.gethostname()
        return [cls(hostname, hostname)]

    @staticmethod
    def get_scheduler_id() -> Optional[int]:
        # The parent shell script is tracked by the local scheduler:
        for parent in psutil.Process().parents():
            cmdline = " ".join(parent.cmdline())
            if "bash" in cmdline and "qlaunch" in cmdline:
                pid = int(parent.pid)
                logger.info(f"Detected scheduled job [{pid}]: {cmdline}")
                return pid
        logger.info("Could not detected scheduler_id of this process.")
        return None



import logging
from typing import Any, Dict

from fastapi import APIRouter, Depends, HTTPException, status
from fastapi.security import OAuth2PasswordRequestForm
from sqlalchemy.orm import Session, exc

from balsam.schemas import UserCreate, UserOut
from balsam.server.models import get_session
from balsam.server.models.crud import users

from .password_utils import verify_password
from .token import create_access_token, user_from_token

logger = logging.getLogger(__name__)

router = APIRouter(prefix="/password")


def authenticate_user_password(db: Session, username: str, password: str) -> UserOut:
    try:
        user = users.get_user_by_username(db, username)
    except exc.NoResultFound:
        logger.info(f"DB does not contain a user with username {username}")
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)

    if not user.hashed_password:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)

    if not verify_password(password, user.hashed_password):
        logger.info(f"User {username} entered a Bad password")
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)

    return UserOut(id=user.id, username=user.username)


@router.post("/login")
def login(form_data: OAuth2PasswordRequestForm = Depends(), db: Session = Depends(get_session)) -> Dict[str, Any]:
    username = form_data.username
    password = form_data.password

    user = authenticate_user_password(db, username, password)
    token, expiry = create_access_token(user)
    return {"access_token": token, "token_type": "bearer", "expiration": expiry}


@router.get("/me", response_model=UserOut)
def profile(user: UserOut = Depends(user_from_token)) -> UserOut:
    return user


@router.post("/register", response_model=UserOut, status_code=status.HTTP_201_CREATED)
def register(user: UserCreate, db: Session = Depends(get_session)) -> UserOut:
    if users.user_exists(db, user.username):
        raise HTTPException(status_code=400, detail="Username already taken")

    new_user = users.create_user(db, user.username, user.password)
    db.commit()
    return new_user



import getpass
import logging
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional

from balsam.schemas import RUNNABLE_STATES, BatchJobState, JobState, SchedulerBackfillWindow

from .service_base import BalsamService

if TYPE_CHECKING:
    from balsam._api.models import BatchJob  # noqa: F401
    from balsam.client import RESTClient

logger = logging.getLogger(__name__)


class ElasticQueueService(BalsamService):
    def __init__(
        self,
        client: "RESTClient",
        site_id: int,
        service_period: int = 60,
        submit_project: str = "datascience",
        submit_queue: str = "balsam",
        job_mode: str = "mpi",
        filter_tags: Optional[Dict[str, str]] = None,
        min_wall_time_min: int = 35,
        max_wall_time_min: int = 360,
        wall_time_pad_min: int = 5,
        min_num_nodes: int = 20,
        max_num_nodes: int = 127,
        max_queue_wait_time_min: int = 10,
        max_queued_jobs: int = 20,
        use_backfill: bool = True,
    ) -> None:
        super().__init__(client=client, service_period=service_period)
        if wall_time_pad_min >= min_wall_time_min:
            raise ValueError("Pad walltime must be less than minimum batch job walltime.")
        self.site_id = site_id
        self.project = submit_project
        self.submit_queue = submit_queue
        self.job_mode = job_mode
        self.filter_tags = filter_tags
        self.max_queue_wait_time_min = max_queue_wait_time_min
        self.min_wall_time_min = min_wall_time_min
        self.max_wall_time_min = max_wall_time_min
        self.wall_time_pad_min = wall_time_pad_min
        self.min_num_nodes = min_num_nodes
        self.max_num_nodes = max_num_nodes
        self.max_queued_jobs = max_queued_jobs
        self.use_backfill = use_backfill
        self.username = getpass.getuser()

    def _get_submission_window(
        self, windows_by_queue: Dict[str, List[SchedulerBackfillWindow]], min_num_nodes: int
    ) -> Optional[SchedulerBackfillWindow]:
        if not self.use_backfill:
            return SchedulerBackfillWindow(num_nodes=self.max_num_nodes, wall_time_min=self.max_wall_time_min)
        windows = windows_by_queue.get(self.submit_queue, [])
        windows = [w for w in windows if w.wall_time_min >= self.min_wall_time_min and w.num_nodes >= min_num_nodes]
        windows = sorted(windows, key=lambda w: w.wall_time_min * w.num_nodes, reverse=True)
        return windows[0] if windows else None

    def get_next_submission(
        self, scheduler_jobs: Iterable["BatchJob"], backfill_windows: Dict[str, List[SchedulerBackfillWindow]]
    ) -> Optional[Dict[str, Any]]:
        Job = self.client.Job
        queued_batchjobs = [j for j in scheduler_jobs if j.state in ("queued", "pending_submission")]
        running_batchjobs = [j for j in scheduler_jobs if j.state == "running"]
        num_reserved_nodes = sum(j.num_nodes for j in queued_batchjobs) + sum(j.num_nodes for j in running_batchjobs)

        logger.debug(
            f"There are {len(queued_batchjobs)} queued, {len(running_batchjobs)} running BatchJobs "
            f"totalling {num_reserved_nodes} nodes."
        )
        if len(queued_batchjobs) + len(running_batchjobs) >= self.max_queued_jobs:
            logger.info(f"At {self.max_queued_jobs} max queued jobs; will not submit")
            return None

        tags = [f"{k}:{v}" for k, v in self.filter_tags.items()] if self.filter_tags else None
        running_jobs = Job.objects.filter(site_id=self.site_id, state=JobState.running, tags=tags)
        runnable_jobs = Job.objects.filter(site_id=self.site_id, state=RUNNABLE_STATES, tags=tags)
        running_num_nodes = sum(float(job.num_nodes) / job.node_packing_count for job in running_jobs)
        runnable_num_nodes = sum(float(job.num_nodes) / job.node_packing_count for job in runnable_jobs)

        logger.debug(
            f"{running_num_nodes} nodes are currently occupied; runnable node footprint is {runnable_num_nodes}"
        )
        if not runnable_jobs:
            logger.info("No Jobs in runnable states; will not submit")
            return None

        window = self._get_submission_window(
            backfill_windows, min_num_nodes=min(job.num_nodes for job in runnable_jobs)
        )
        logger.debug(f"Largest node window: {window}")

        if not window:
            logger.info("No eligible backfill windows; will not submit")
            return None

        # The number of nodes currently or soon to be allocated, minus the footprint of currently running jobs
        idle_node_count = num_reserved_nodes - running_num_nodes

        if runnable_num_nodes > idle_node_count:
            request_num_nodes = max(self.min_num_nodes, runnable_num_nodes - idle_node_count)
            request_num_nodes = min(
                request_num_nodes,
                window.num_nodes,
                self.max_num_nodes,
            )
            return {
                "site_id": self.site_id,
                "project": self.project,
                "queue": self.submit_queue,
                "job_mode": self.job_mode,
                "filter_tags": self.filter_tags,
                "num_nodes": request_num_nodes,
                "wall_time_min": window.wall_time_min - self.wall_time_pad_min,
            }
        logger.info("Idle node count meets or exceeds runnable footprint; no need to submit.")
        return None

    def run_cycle(self) -> None:
        BatchJob = self.client.BatchJob  # noqa: F811
        site = self.client.Site.objects.get(id=self.site_id)
        tags = [f"{k}:{v}" for k, v in self.filter_tags.items()] if self.filter_tags else None
        backfill_windows = site.backfill_windows
        scheduler_jobs = list(
            BatchJob.objects.filter(site_id=self.site_id, filter_tags=tags, queue=self.submit_queue)
        )

        sub = self.get_next_submission(scheduler_jobs, backfill_windows)
        if sub:
            new_job = BatchJob(**sub)
            new_job.save()
            logger.info(f"Submitted new BatchJob: {new_job}")

        cancel_jobs = [
            job
            for job in scheduler_jobs
            if job.state == "queued"
            and job.scheduler_id in site.queued_jobs
            and site.queued_jobs[job.scheduler_id].queued_time_min > self.max_queue_wait_time_min
        ]
        for job in cancel_jobs:
            try:
                batch_job = BatchJob.objects.get(scheduler_id=job.scheduler_id, site_id=self.site_id)
            except BatchJob.DoesNotExist:
                logger.warning(
                    f"Trying to delete BatchJob with scheduler id {job.scheduler_id}, but it does not exist in the API."
                )
            else:
                if batch_job.state != BatchJobState.pending_deletion:
                    batch_job.state = BatchJobState.pending_deletion
                    batch_job.save()
                    logger.info(f"Marked queued BatchJob {job.scheduler_id} for deletion: exceed max queue wait time")

    def cleanup(self) -> None:
        logger.info("Exiting ElasticQueue service")



import json
import logging
import os
from abc import ABCMeta
from datetime import datetime
from importlib import import_module
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type, Union, cast
from uuid import UUID

import yaml
from pydantic import AnyUrl, BaseSettings, Field, ValidationError, validator

from balsam.client import NotAuthenticatedError, RequestsClient
from balsam.platform.app_run import AppRun
from balsam.platform.compute_node import ComputeNode
from balsam.platform.scheduler import SchedulerInterface
from balsam.platform.transfer import GlobusTransferInterface, TransferInterface
from balsam.schemas import AllowedQueue
from balsam.util import config_file_logging

if TYPE_CHECKING:
    from balsam.site.service.service_base import BalsamService

logger = logging.getLogger(__name__)


class InvalidSettings(Exception):
    pass


def balsam_home() -> Path:
    return Path.home().joinpath(".balsam")


def get_class_path(cls: type) -> str:
    return cls.__module__ + "." + cls.__name__


def import_string(dotted_path: str) -> Any:
    """
    Stolen from pydantic. Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import fails.
    """
    try:
        module_path, class_name = dotted_path.strip(" ").rsplit(".", 1)
    except ValueError as e:
        raise ImportError(f'"{dotted_path}" doesn\'t look like a module path') from e

    module = import_module(module_path)
    try:
        return getattr(module, class_name)
    except AttributeError as e:
        raise ImportError(f'Module "{module_path}" does not define a "{class_name}" attribute') from e


class ClientSettings(BaseSettings):
    api_root: str
    client_class: Type[RequestsClient] = Field("balsam.client.BasicAuthRequestsClient")
    token: Optional[str] = None
    token_expiry: Optional[datetime] = None
    connect_timeout: float = 6.2
    read_timeout: float = 120.0
    retry_count: int = 10

    @validator("client_class", pre=True, always=True)
    def load_client_class(cls, v: str) -> Type[RequestsClient]:
        loaded = import_string(v) if isinstance(v, str) else v
        if not issubclass(loaded, RequestsClient):
            raise TypeError(f"client_class must subclass {get_class_path(RequestsClient)}")
        return cast(Type[RequestsClient], loaded)

    @staticmethod
    def settings_path() -> Path:
        env_path = os.environ.get("BALSAM_CLIENT_PATH")
        if env_path:
            return Path(env_path)
        else:
            return balsam_home().joinpath("client.yml")

    @classmethod
    def load_from_file(cls) -> "ClientSettings":
        try:
            with open(cls.settings_path()) as fp:
                data = yaml.safe_load(fp)
        except FileNotFoundError:
            raise NotAuthenticatedError(
                f"Client credentials {cls.settings_path()} do not exist. " f"Please authenticate with `balsam login`."
            )
        return cls(**data)

    def save_to_file(self) -> None:
        data = self.dict()
        cls = data["client_class"]
        data["client_class"] = get_class_path(cls)

        settings_path = self.settings_path()
        if not settings_path.parent.is_dir():
            settings_path.parent.mkdir()
            settings_path.parent.chmod(0o700)
        if settings_path.exists():
            os.chmod(settings_path, 0o600)
        else:
            open(settings_path, "w").close()
            os.chmod(settings_path, 0o600)
        with open(settings_path, "w+") as fp:
            yaml.dump(data, fp, sort_keys=False, indent=4)

    def build_client(self) -> RequestsClient:
        client = self.client_class(**self.dict(exclude={"client_class"}))
        return client


class LoggingConfig(BaseSettings):
    level: str = "DEBUG"
    format: str = "%(asctime)s.%(msecs)03d | %(process)d | %(levelname)s | %(name)s:%(lineno)s] %(message)s"
    datefmt: str = "%Y-%m-%d %H:%M:%S"
    buffer_num_records: int = 1024
    flush_period: int = 30


class SchedulerSettings(BaseSettings):
    scheduler_class: Type[SchedulerInterface] = Field("balsam.platform.scheduler.CobaltScheduler")
    sync_period: int = 60
    allowed_queues: Dict[str, AllowedQueue] = {
        "default": AllowedQueue(max_nodes=4010, max_walltime=24 * 60, max_queued_jobs=20),
        "debug-cache-quad": AllowedQueue(max_nodes=8, max_walltime=60, max_queued_jobs=1),
    }
    allowed_projects: List[str] = ["datascience", "magstructsADSP"]
    optional_batch_job_params: Dict[str, str] = {"singularity_prime_cache": "no"}
    job_template_path: Path = Path("job-template.sh")

    @validator("scheduler_class", pre=True, always=True)
    def load_scheduler_class(cls, v: str) -> Type[SchedulerInterface]:
        loaded = import_string(v) if isinstance(v, str) else v
        if not issubclass(loaded, SchedulerInterface):
            raise TypeError(f"scheduler_class must subclass {get_class_path(SchedulerInterface)}")
        return cast(Type[SchedulerInterface], loaded)


class QueueMaintainerSettings(BaseSettings):
    submit_period: int = 60
    submit_project: str = "local"
    submit_queue: str = "local"
    job_mode: str = "mpi"
    num_queued_jobs: int = 5
    num_nodes: int = 1
    wall_time_min: int = 1


class ElasticQueueSettings(BaseSettings):
    service_period: int = 60
    submit_project: str = "datascience"
    submit_queue: str = "balsam"
    job_mode: str = "mpi"
    min_wall_time_min: int = 35
    max_wall_time_min: int = 360
    wall_time_pad_min: int = 5
    min_num_nodes: int = 20
    max_num_nodes: int = 127
    max_queue_wait_time_min: int = 10
    max_queued_jobs: int = 20
    use_backfill: bool = True


class FileCleanerSettings(BaseSettings):
    cleanup_batch_size: int = 180
    service_period: int = 30


class ProcessingSettings(BaseSettings):
    num_workers: int = 5
    prefetch_depth: int = 1000


class TransferSettings(BaseSettings):
    transfer_locations: Dict[str, str] = {"theta_dtn": "globus://08925f04-569f-11e7-bef8-22000b9a448b"}
    max_concurrent_transfers: int = 5
    globus_endpoint_id: Optional[UUID] = None
    transfer_batch_size: int = 100
    num_items_query_limit: int = 2000
    service_period: int = 5


class LauncherSettings(BaseSettings):
    idle_ttl_sec: int = 60
    delay_sec: int = 1
    error_tail_num_lines: int = 10
    max_concurrent_mpiruns: int = 1000
    compute_node: Type[ComputeNode] = Field("balsam.platform.compute_node.ThetaKNLNode")
    mpi_app_launcher: Type[AppRun] = Field("balsam.platform.app_run.ThetaAprun")
    local_app_launcher: Type[AppRun] = Field("balsam.platform.app_run.LocalAppRun")
    mpirun_allows_node_packing: bool = False
    serial_mode_prefetch_per_rank: int = 64
    serial_mode_startup_params: Dict[str, str] = {"cpu_affinity": "none"}

    @validator("compute_node", pre=True, always=True)
    def load_compute_node_class(cls, v: str) -> Type[ComputeNode]:
        loaded = import_string(v) if isinstance(v, str) else v
        if not issubclass(loaded, ComputeNode):
            raise TypeError(f"compute_node must subclass {get_class_path(ComputeNode)}")
        return cast(Type[ComputeNode], loaded)

    @validator("mpi_app_launcher", pre=True, always=True)
    def load_mpi_app_launcher(cls, v: str) -> Type[AppRun]:
        loaded = import_string(v) if isinstance(v, str) else v
        if not issubclass(loaded, AppRun):
            raise TypeError(f"mpi_app_launcher must subclass {get_class_path(AppRun)}")
        return cast(Type[AppRun], loaded)

    @validator("local_app_launcher", pre=True, always=True)
    def load_local_app_launcher(cls, v: str) -> Type[AppRun]:
        loaded = import_string(v) if isinstance(v, str) else v
        if not issubclass(loaded, AppRun):
            raise TypeError(f"local_app_launcher must subclass {get_class_path(AppRun)}")
        return cast(Type[AppRun], loaded)


class Settings(BaseSettings):
    logging: LoggingConfig = LoggingConfig()
    filter_tags: Dict[str, str] = {"workflow": "test-1", "system": "H2O"}

    # Balsam service modules
    launcher: LauncherSettings = LauncherSettings()
    scheduler: Optional[SchedulerSettings] = SchedulerSettings()
    processing: Optional[ProcessingSettings] = ProcessingSettings()
    transfers: Optional[TransferSettings] = TransferSettings()
    file_cleaner: Optional[FileCleanerSettings] = FileCleanerSettings()
    queue_maintainer: Optional[QueueMaintainerSettings] = QueueMaintainerSettings()
    elastic_queue: Optional[ElasticQueueSettings] = ElasticQueueSettings()

    def save(self, path: Union[str, Path]) -> None:
        with open(path, "w") as fp:
            fp.write(self.dump_yaml())

    def dump_yaml(self) -> str:
        return cast(
            str,
            yaml.dump(
                json.loads(self.json()),
                sort_keys=False,
                indent=4,
            ),
        )

    @classmethod
    def load(cls, path: Union[str, Path]) -> "Settings":
        with open(path) as fp:
            raw_data = yaml.safe_load(fp)
        return cls(**raw_data)

    class Config:
        json_encoders = {
            type: get_class_path,
            ABCMeta: get_class_path,
        }


class SiteConfig:
    """
    Uses above settings to build components and provide dependencies
    No component should refer to external settings or set its own dependencies
    Instead, this class builds and injects needed settings/dependencies at runtime
    """

    def __init__(self, site_path: Union[str, Path, None] = None, settings: Optional[Settings] = None) -> None:
        site_path, site_id = self.resolve_site_path(site_path)
        self.site_path: Path = site_path
        self.site_id: int = site_id
        self.client = ClientSettings.load_from_file().build_client()

        if settings is not None:
            if not isinstance(settings, Settings):
                raise ValueError(
                    "If you're passing the settings kwarg, it must be an instance of balsam.config.Settings. "
                    "Otherwise, leave settings=None to auto-load the settings stored at BALSAM_SITE_PATH."
                )
            self.settings = settings
            return

        yaml_settings = self.site_path.joinpath("settings.yml")

        if not yaml_settings.is_file():
            raise FileNotFoundError(f"{site_path} must contain a settings.yml")
        try:
            self.settings = Settings.load(yaml_settings)
        except ValidationError as exc:
            raise InvalidSettings(f"{yaml_settings} is invalid:\n{exc}")

    def build_services(self) -> "List[BalsamService]":
        from balsam.site.service import (
            ElasticQueueService,
            FileCleanerService,
            ProcessingService,
            QueueMaintainerService,
            SchedulerService,
            TransferService,
        )

        services: List[BalsamService] = []

        if self.settings.scheduler:
            scheduler_service = SchedulerService(
                client=self.client,
                site_id=self.site_id,
                submit_directory=self.job_path,
                filter_tags=self.settings.filter_tags,
                **dict(self.settings.scheduler),  # does not convert sub-models to dicts
            )
            services.append(scheduler_service)

        if self.settings.queue_maintainer:
            queue_maintainer = QueueMaintainerService(
                client=self.client,
                site_id=self.site_id,
                filter_tags=self.settings.filter_tags,
                **dict(self.settings.queue_maintainer),  # does not convert sub-models to dicts
            )
            services.append(queue_maintainer)

        if self.settings.elastic_queue:
            elastic_queue = ElasticQueueService(
                client=self.client,
                site_id=self.site_id,
                filter_tags=self.settings.filter_tags,
                **dict(self.settings.elastic_queue),
            )
            services.append(elastic_queue)

        if self.settings.processing:
            processing_service = ProcessingService(
                client=self.client,
                site_id=self.site_id,
                data_path=self.data_path,
                apps_path=self.apps_path,
                filter_tags=self.settings.filter_tags,
                **dict(self.settings.processing),  # does not convert sub-models to dicts
            )
            services.append(cast("BalsamService", processing_service))

        if self.settings.transfers:
            transfer_settings = dict(self.settings.transfers)
            transfer_interfaces: Dict[str, TransferInterface] = {}
            endpoint_id = transfer_settings.pop("globus_endpoint_id")
            if endpoint_id:
                transfer_interfaces["globus"] = GlobusTransferInterface(endpoint_id)
            transfer_service = TransferService(
                client=self.client,
                site_id=self.site_id,
                data_path=self.data_path,
                transfer_interfaces=transfer_interfaces,
                **dict(transfer_settings),
            )
            services.append(transfer_service)

        if self.settings.file_cleaner:
            cleaner_service = FileCleanerService(
                client=self.client,
                site_id=self.site_id,
                apps_path=self.apps_path,
                data_path=self.data_path,
                **dict(self.settings.file_cleaner),
            )
            services.append(cleaner_service)

        return services

    @staticmethod
    def resolve_site_path(site_path: Union[None, str, Path] = None) -> Tuple[Path, int]:
        # Site determined from either passed argument, environ,
        # or walking up parent directories, in that order
        site_path = site_path or os.environ.get("BALSAM_SITE_PATH") or SiteConfig.search_site_dir()
        if site_path is None:
            raise ValueError(
                "Initialize SiteConfig with a `site_path` or set env BALSAM_SITE_PATH "
                "to a Balsam site directory containing a settings.py file."
            )

        site_path = Path(site_path).resolve()
        if not site_path.is_dir():
            raise FileNotFoundError(f"BALSAM_SITE_PATH {site_path} must point to an existing Balsam site directory")
        try:
            site_id = int(site_path.joinpath(".balsam-site").read_text())
        except FileNotFoundError:
            raise FileNotFoundError(
                f"BALSAM_SITE_PATH {site_path} is not a valid Balsam site directory "
                f"(does not contain a .balsam-site file)"
            )
        os.environ["BALSAM_SITE_PATH"] = str(site_path)
        return site_path, site_id

    @staticmethod
    def search_site_dir() -> Optional[Path]:
        check_dir = Path.cwd()
        while check_dir.as_posix() != "/":
            if check_dir.joinpath(".balsam-site").is_file():
                return check_dir
            check_dir = check_dir.parent
        return None

    @property
    def apps_path(self) -> Path:
        return self.site_path.joinpath("apps")

    @property
    def log_path(self) -> Path:
        return self.site_path.joinpath("log")

    @property
    def job_path(self) -> Path:
        return self.site_path.joinpath("qsubmit")

    @property
    def data_path(self) -> Path:
        return self.site_path.joinpath("data")

    def enable_logging(self, basename: str, filename: Optional[str] = None) -> Dict[str, Any]:
        if filename is None:
            ts = datetime.now().strftime("%Y-%m-%d_%H%M%S")
            filename = f"{basename}_{ts}.log"
        log_path = self.log_path.joinpath(filename)
        config_file_logging(
            filename=log_path,
            **self.settings.logging.dict(),
        )
        return {"filename": log_path, **self.settings.logging.dict()}

    def update_site_from_config(self) -> None:
        site = self.client.Site.objects.get(id=self.site_id)
        old_dict = site.display_dict()
        if self.settings.scheduler:
            site.allowed_projects = self.settings.scheduler.allowed_projects
            site.allowed_queues = self.settings.scheduler.allowed_queues
            site.optional_batch_job_params = self.settings.scheduler.optional_batch_job_params
        if self.settings.transfers:
            site.transfer_locations = cast(Dict[str, AnyUrl], self.settings.transfers.transfer_locations)
            site.globus_endpoint_id = self.settings.transfers.globus_endpoint_id

        new_dict = site.display_dict()
        diff = {k: (old_dict[k], new_dict[k]) for k in old_dict if old_dict[k] != new_dict[k]}
        if diff:
            site.save()
            diff_str = "\n".join(f"{k}={diff[k][0]} --> {diff[k][1]}" for k in diff)
            logger.info(f"Updated Site parameters:\n{diff_str}")



import logging
import time
from typing import Iterator

logger = logging.getLogger(__name__)


def countdown_timer_min(time_limit_min: int, delay_sec: int) -> Iterator[float]:
    start = time.time()
    next_time = start + delay_sec
    while True:
        now = time.time()
        to_sleep = next_time - now
        if to_sleep <= 0:
            next_time = now + delay_sec
        else:
            time.sleep(to_sleep)
            now += to_sleep
            next_time = now + delay_sec
        elapsed_min = (now - start) / 60.0
        remaining_min = time_limit_min - elapsed_min

        if remaining_min > 0:
            yield remaining_min
        else:
            return



import logging
import os
import random
import time
from json import JSONDecodeError
from pprint import pformat
from typing import Any, Dict, List, Optional, Type, Union

import requests

from . import urls
from .rest_base_client import RESTClient

logger = logging.getLogger(__name__)

OptionalAnyJSON = Optional[Union[Dict[str, Any], List[Any]]]


class NotAuthenticatedError(Exception):
    pass


class RequestsClient(RESTClient):

    _client_classes: "Dict[str, Type[RequestsClient]]" = {}

    @staticmethod
    def discover_supported_client(base_url: str) -> "RequestsClient":
        url = base_url.rstrip("/") + "/" + urls.CHECK_LOGIN_FLOWS
        auth_methods: List[str] = requests.get(url).json()

        if "LoginMethod.oauth_device" in auth_methods:
            cls = RequestsClient._client_classes["oauth_device"]
        elif "LoginMethod.password" in auth_methods:
            cls = RequestsClient._client_classes["password"]
        else:
            raise NotImplementedError(f"This client does not support the server's auth methods: {auth_methods}")

        return cls(api_root=base_url)

    def __init__(
        self, api_root: str, connect_timeout: float = 30.1, read_timeout: float = 120.0, retry_count: int = 10
    ) -> None:
        self.api_root = api_root
        self.connect_timeout = connect_timeout
        self.read_timeout = read_timeout
        self.retry_count = retry_count
        self._session: Optional[requests.Session] = None
        self._pid = os.getpid()
        self._authenticated = False
        self.token: Optional[str] = None
        self._attempt: int = 0

    @property
    def session(self) -> requests.Session:
        """
        !!! WARNING !!!
        requests.Session is not multiprocessing-safe. You will get
        crossed-wires and mixed up responses; leading to strange and
        near-impossible-to-debug issues. As an extra precaution here we start
        a new Session if a PID change is detected.
        """
        pid = os.getpid()
        if pid != self._pid or self._session is None:
            self._session = requests.Session()
            self._pid = pid
            if self.token:
                self._session.headers["Authorization"] = f"Bearer {self.token}"
        return self._session

    def close_session(self) -> None:
        self._session = None
        return None

    def backoff(self, reason: Exception) -> None:
        if self._attempt > self.retry_count:
            raise TimeoutError(f"Exceeded max retries: {reason}")
        sleep_time = 2 ** self._attempt + random.random()
        time.sleep(sleep_time)
        self._attempt += 1

    def request(
        self,
        url: str,
        http_method: str,
        params: Optional[Dict[str, Any]] = None,
        json: OptionalAnyJSON = None,
        data: OptionalAnyJSON = None,
        authenticating: bool = False,
    ) -> OptionalAnyJSON:
        if not self._authenticated and not authenticating:
            raise NotAuthenticatedError("Cannot perform unauthenticated request. Please login with `balsam login`")
        absolute_url = self.api_root.rstrip("/") + "/" + url.lstrip("/")
        self._attempt = 0
        while True:
            try:
                logger.debug(f"{http_method}: {absolute_url} {params if params else ''}")
                response = self._do_request(absolute_url, http_method, params, json, data)
            except requests.Timeout as exc:
                logger.warning(f"Attempt Retry of Timed-out request {http_method} {absolute_url}")
                self.backoff(exc)
            except requests.ConnectionError as exc:
                logger.warning(f"Attempt retry ({self._attempt} of {self.retry_count}) of connection: {exc}")
                self.backoff(exc)
            else:
                try:
                    return response.json()  # type: ignore
                except (ValueError, JSONDecodeError):
                    if http_method != "DELETE":
                        raise
                    return None

    def _do_request(
        self,
        absolute_url: str,
        http_method: str,
        params: Optional[Dict[str, Any]],
        json: OptionalAnyJSON,
        data: OptionalAnyJSON,
    ) -> requests.Response:
        response = self.session.request(
            http_method,
            url=absolute_url,
            params=params,
            json=json,
            data=data,
            timeout=(self.connect_timeout, self.read_timeout),
        )
        if response.status_code >= 400:
            self._raise_with_explanation(response)
        return response

    def _raise_with_explanation(self, response: requests.Response) -> None:
        """
        Add the API's informative error message to Requests' generic status Exception
        """
        try:
            explanation = response.json()
        except JSONDecodeError:
            explanation = ""
        else:
            explanation = "\n" + pformat(explanation, indent=4)

        if response.reason is None:
            response.reason = explanation
        elif isinstance(response.reason, bytes):
            response.reason = response.reason.decode("utf-8") + explanation
        else:
            response.reason += explanation
        response.raise_for_status()



from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field, validator

MAX_APP_SERIALIZED_SIZE = 262_144


class AppParameter(BaseModel):
    required: bool
    default: Optional[str]
    help: str = ""

    @validator("default")
    def is_required_or_has_default(cls, default_value: Optional[str], values: Dict[str, Any]) -> Optional[str]:
        if values.get("required", True):
            if default_value:
                raise ValueError("cannot be required and have default")
        else:
            if not default_value:
                raise ValueError("optional param needs a default")
        return default_value


class TransferDirection(str, Enum):
    stage_in = "in"
    stage_out = "out"


class TransferSlot(BaseModel):
    required: bool
    direction: TransferDirection
    local_path: Path
    description: str = ""
    recursive: bool = False

    @validator("local_path")
    def path_is_relative(cls, v: Path) -> Path:
        if v.is_absolute():
            raise ValueError("Cannot use absolute path")
        return v


class AppBase(BaseModel):
    site_id: int = Field(..., example=3, description="Site id at which this App is registered")
    name: str = Field(..., example="NWChemGeomOpt", description="Python AppDef class name")
    serialized_class: str = Field(..., description="Base64-encoded ApplicationDefinition payload")
    source_code: str = Field(..., description="App-introspected source code")
    description: str = Field("", example="NWChem7 geometry optimizer", description="The App class docstring")
    parameters: Dict[str, AppParameter] = Field(
        {},
        example={
            "input_file": {
                "required": False,
                "default": "input.nw",
                "help": "Path to input deck",
            }
        },
        description="Allowed parameters in the App command",
    )
    transfers: Dict[str, TransferSlot] = Field(
        {},
        example={
            "input_file": {
                "required": True,
                "direction": "in",
                "local_path": "input.nw",
                "description": "Input Deck",
                "recursive": "False",
            }
        },
        description="Allowed transfer slots in the App",
    )

    @validator("name")
    def is_valid_class_name(cls, v: str) -> str:
        if not v.isidentifier():
            raise ValueError(f"{v} is not a valid class name")
        return v

    @validator("serialized_class")
    def max_class_len(cls, v: str) -> str:
        if len(v) > MAX_APP_SERIALIZED_SIZE:
            raise AssertionError(f"Serialized App cannot be larger than {MAX_APP_SERIALIZED_SIZE}")
        return v


class AppCreate(AppBase):
    pass


class AppUpdate(AppBase):
    site_id: int = Field(None, example=3, description="Site id at which this App is registered")
    name: str = Field(None, example="NWChem7GeomOpt", description="Python AppDef class name")
    serialized_class: str = Field(None, description="Base64-encoded ApplicationDefinition payload")
    source_code: str = Field(None, description="App-introspected source code")


class AppOut(AppBase):
    class Config:
        orm_mode = True

    id: int = Field(..., example=234)


class PaginatedAppsOut(BaseModel):
    count: int
    results: List[AppOut]
